import pytest

import graphene
from graphene.relay import Node
from graphene_django import DjangoConnectionField, DjangoObjectType
from graphene_django.tests.models import Article, Reporter
from graphene_django.utils import DJANGO_FILTER_INSTALLED

pytestmark = []

if DJANGO_FILTER_INSTALLED:
    from graphene_django.filter import DjangoFilterConnectionField
else:
    pytestmark.append(
        pytest.mark.skipif(
            True, reason="django_filters not installed or not compatible"
        )
    )


@pytest.fixture
def schema():
    class ReporterType(DjangoObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)
            fields = "__all__"

    class ArticleType(DjangoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)
            fields = "__all__"
            filter_fields = {
                "lang": ["exact", "in"],
                "reporter__a_choice": ["exact", "in"],
            }

    class Query(graphene.ObjectType):
        all_reporters = DjangoConnectionField(ReporterType)
        all_articles = DjangoFilterConnectionField(ArticleType)

    schema = graphene.Schema(query=Query)
    return schema


@pytest.fixture
def reporter_article_data():
    john = Reporter.objects.create(
        first_name="John", last_name="Doe", email="johndoe@example.com", a_choice=1
    )
    jane = Reporter.objects.create(
        first_name="Jane", last_name="Doe", email="janedoe@example.com", a_choice=2
    )
    Article.objects.create(
        headline="Article Node 1", reporter=john, editor=john, lang="es"
    )
    Article.objects.create(
        headline="Article Node 2", reporter=john, editor=john, lang="en"
    )
    Article.objects.create(
        headline="Article Node 3", reporter=jane, editor=jane, lang="en"
    )


def test_filter_enum_on_connection(schema, reporter_article_data):
    """
    Check that we can filter with enums on a connection.
    """
    query = """
        query {
            allArticles(lang: ES) {
                edges {
                    node {
                        headline
                    }
                }
            }
        }
    """

    expected = {
        "allArticles": {
            "edges": [
                {"node": {"headline": "Article Node 1"}},
            ]
        }
    }

    result = schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_filter_on_foreign_key_enum_field(schema, reporter_article_data):
    """
    Check that we can filter with enums on a field from a foreign key.
    """
    query = """
        query {
            allArticles(reporter_AChoice: A_1) {
                edges {
                    node {
                        headline
                    }
                }
            }
        }
    """

    expected = {
        "allArticles": {
            "edges": [
                {"node": {"headline": "Article Node 1"}},
                {"node": {"headline": "Article Node 2"}},
            ]
        }
    }

    result = schema.execute(query)
    assert not result.errors
    assert result.data == expected


def test_filter_enum_field_schema_type(schema):
    """
    Check that the type in the filter is an enum like on the object type.
    """
    schema_str = str(schema)

    assert (
        '''type ArticleType implements Node {
  """The ID of the object"""
  id: ID!
  headline: String!
  pubDate: Date!
  pubDateTime: DateTime!
  reporter: ReporterType!
  editor: ReporterType!

  """Language"""
  lang: TestsArticleLangChoices!
  importance: TestsArticleImportanceChoices
}'''
        in schema_str
    )

    filters = {
        "offset": "Int",
        "before": "String",
        "after": "String",
        "first": "Int",
        "last": "Int",
        "lang": "TestsArticleLangChoices",
        "lang_In": "[TestsArticleLangChoices]",
        "reporter_AChoice": "TestsReporterAChoiceChoices",
        "reporter_AChoice_In": "[TestsReporterAChoiceChoices]",
    }
    filters_str = ", ".join(
        [f"{filter_field}: {gql_type}" for filter_field, gql_type in filters.items()]
    )
    assert f"  allArticles({filters_str}): ArticleTypeConnection\n" in schema_str



import graphene
from graphene import Schema, relay

from ..types import DjangoObjectType
from .models import Article, Reporter


class Character(DjangoObjectType):
    class Meta:
        model = Reporter
        interfaces = (relay.Node,)
        fields = "__all__"

    def get_node(self, info, id):
        pass


class Human(DjangoObjectType):
    raises = graphene.String()

    class Meta:
        model = Article
        interfaces = (relay.Node,)
        fields = "__all__"

    def resolve_raises(self, info):
        raise Exception("This field should raise exception")

    def get_node(self, info, id):
        pass


class Query(graphene.ObjectType):
    human = graphene.Field(Human)

    def resolve_human(self, info):
        return Human()


schema = Schema(query=Query)



import sys
from pathlib import PurePath

# For backwards compatibility, we import JSONField to have it available for import via
# this compat module (https://github.com/graphql-python/graphene-django/issues/1428).
# Django's JSONField is available in Django 3.2+ (the minimum version we support)
from django.db.models import JSONField


class MissingType:
    def __init__(self, *args, **kwargs):
        pass


try:
    # Postgres fields are only available in Django with psycopg2 installed
    # and we cannot have psycopg2 on PyPy
    from django.contrib.postgres.fields import (
        ArrayField,
        HStoreField,
        IntegerRangeField,
        RangeField,
    )
except ImportError:
    IntegerRangeField, HStoreField, RangeField = (MissingType,) * 3

    # For unit tests we fake ArrayField using JSONFields
    if any(
        PurePath(sys.argv[0]).match(p)
        for p in [
            "**/pytest",
            "**/py.test",
            "**/pytest/__main__.py",
        ]
    ):

        class ArrayField(JSONField):
            def __init__(self, *args, **kwargs):
                if len(args) > 0:
                    self.base_field = args[0]
                super().__init__(**kwargs)

    else:
        ArrayField = MissingType



from django_filters.filters import FilterMethod

from .typed_filter import TypedFilter


class ListFilterMethod(FilterMethod):
    def __call__(self, qs, value):
        if value is None:
            return qs
        return self.method(qs, self.f.field_name, value)


class ListFilter(TypedFilter):
    """
    Filter that takes a list of value as input.
    It is for example used for `__in` filters.
    """

    @TypedFilter.method.setter
    def method(self, value):
        """
        Override method setter so that in case a custom `method` is provided
        (see documentation https://django-filter.readthedocs.io/en/stable/ref/filters.html#method),
        it doesn't fall back to checking if the value is in `EMPTY_VALUES` (from the `__call__` method
        of the `FilterMethod` class) and instead use our ListFilterMethod that consider empty lists as values.

        Indeed when providing a `method` the `filter` method below is overridden and replaced by `FilterMethod(self)`
        which means that the validation of the empty value is made by the `FilterMethod.__call__` method instead.
        """
        TypedFilter.method.fset(self, value)
        if value is not None:
            self.filter = ListFilterMethod(self)

    def filter(self, qs, value):
        """
        Override the default filter class to check first whether the list is
        empty or not.
        This needs to be done as in this case we expect to get an empty output
        (if not an exclude filter) but django_filter consider an empty list
        to be an empty input value (see `EMPTY_VALUES`) meaning that
        the filter does not need to be applied (hence returning the original
        queryset).
        """
        if value is not None and len(value) == 0:
            if self.exclude:
                return qs
            else:
                return qs.none()
        else:
            return super().filter(qs, value)



import pytest

from ...compat import ArrayField, MissingType


@pytest.mark.skipif(ArrayField is MissingType, reason="ArrayField should exist")
def test_array_field_len_filter(schema):
    query = """
    query {
        events (tags_Len: 2) {
            edges {
                node {
                    name
                }
            }
        }
    }
    """
    result = schema.execute(query)
    assert not result.errors
    assert result.data["events"]["edges"] == [
        {"node": {"name": "Musical"}},
        {"node": {"name": "Ballet"}},
    ]

    query = """
    query {
        events (tags_Len: 0) {
            edges {
                node {
                    name
                }
            }
        }
    }
    """
    result = schema.execute(query)
    assert not result.errors
    assert result.data["events"]["edges"] == [
        {"node": {"name": "Speech"}},
    ]

    query = """
    query {
        events (tags_Len: 10) {
            edges {
                node {
                    name
                }
            }
        }
    }
    """
    result = schema.execute(query)
    assert not result.errors
    assert result.data["events"]["edges"] == []

    query = """
    query {
        events (tags_Len: "2") {
            edges {
                node {
                    name
                }
            }
        }
    }
    """
    result = schema.execute(query)
    assert len(result.errors) == 1
    assert result.errors[0].message == 'Int cannot represent non-integer value: "2"'

    query = """
    query {
        events (tags_Len: True) {
            edges {
                node {
                    name
                }
            }
        }
    }
    """
    result = schema.execute(query)
    assert len(result.errors) == 1
    assert result.errors[0].message == "Int cannot represent non-integer value: True"


@pytest.mark.skipif(ArrayField is MissingType, reason="ArrayField should exist")
def test_array_field_custom_filter(schema):
    query = """
    query {
        events (tags_Len_In: 2) {
            edges {
                node {
                    name
                }
            }
        }
    }
    """
    result = schema.execute(query)
    assert not result.errors
    assert result.data["events"]["edges"] == [
        {"node": {"name": "Ballet"}},
        {"node": {"name": "Musical"}},
    ]

    query = """
    query {
        events (tags_Len_In: [0, 2]) {
            edges {
                node {
                    name
                }
            }
        }
    }
    """
    result = schema.execute(query)
    assert not result.errors
    assert result.data["events"]["edges"] == [
        {"node": {"name": "Ballet"}},
        {"node": {"name": "Musical"}},
        {"node": {"name": "Speech"}},
    ]

    query = """
    query {
        events (tags_Len_In: [10]) {
            edges {
                node {
                    name
                }
            }
        }
    }
    """
    result = schema.execute(query)
    assert not result.errors
    assert result.data["events"]["edges"] == []

    query = """
    query {
        events (tags_Len_In: []) {
            edges {
                node {
                    name
                }
            }
        }
    }
    """
    result = schema.execute(query)
    assert not result.errors
    assert result.data["events"]["edges"] == []

    query = """
    query {
        events (tags_Len_In: "12") {
            edges {
                node {
                    name
                }
            }
        }
    }
    """
    result = schema.execute(query)
    assert len(result.errors) == 1
    assert result.errors[0].message == 'Int cannot represent non-integer value: "12"'

    query = """
    query {
        events (tags_Len_In: True) {
            edges {
                node {
                    name
                }
            }
        }
    }
    """
    result = schema.execute(query)
    assert len(result.errors) == 1
    assert result.errors[0].message == "Int cannot represent non-integer value: True"



from .testing import GraphQLTestCase
from .utils import (
    DJANGO_FILTER_INSTALLED,
    bypass_get_queryset,
    camelize,
    get_model_fields,
    get_reverse_fields,
    is_valid_django_model,
    maybe_queryset,
)

__all__ = [
    "DJANGO_FILTER_INSTALLED",
    "get_reverse_fields",
    "maybe_queryset",
    "get_model_fields",
    "camelize",
    "is_valid_django_model",
    "GraphQLTestCase",
    "bypass_get_queryset",
]



from graphene import ObjectType, String


class DjangoDebugException(ObjectType):
    class Meta:
        description = "Represents a single exception raised."

    exc_type = String(required=True, description="The class of the exception")
    message = String(required=True, description="The message of the exception")
    stack = String(required=True, description="The stack trace")
